/**************************************************************************
 * Copyright(c) 1998-1999, ALICE Experiment at CERN, All rights reserved. *
 *                                                                        *
 * Author: The ALICE Off-line Project.                                    *
 * Contributors are mentioned in the code where appropriate.              *
 *                                                                        *
 * Permission to use, copy, modify and distribute this software and its   *
 * documentation strictly for non-commercial purposes is hereby granted   *
 * without fee, provided that the above copyright notice appears in all   *
 * copies and that both the copyright notice and this permission notice   *
 * appear in the supporting documentation. The authors make no claims     *
 * about the suitability of this software for any purpose. It is          *
 * provided "as is" without express or implied warranty.                  *
 **************************************************************************/
 
//-------------------------------------------------------------------------
//                      Class AliAODSimpleEvent
//  Simple collection of reconstructed tracks, selected from an ESD event
// 
// author: A. Pulvirenti             (email: alberto.pulvirenti@ct.infn.it)
//-------------------------------------------------------------------------

#include <Riostream.h>

#include <TTree.h>
#include <TString.h>
#include <TParticle.h>
#include <TObjArray.h>
#include <TRefArray.h>
#include <TClonesArray.h>

#include "AliESD.h"
#include "AliStack.h"
#include "AliESDtrack.h"
#include "AliAODSimpleTrack.h"
#include "AliAODSimpleEvent.h"

ClassImp(AliAODSimpleEvent)

TClonesArray* AliAODSimpleEvent::fgTracks = 0;
TArrayD       AliAODSimpleEvent::fgPrior(kParticleTypes);
TArrayD       AliAODSimpleEvent::fgDefMass(kArrayIndexes);
TArrayI       AliAODSimpleEvent::fgDefPDG(kParticleTypes);
Bool_t        AliAODSimpleEvent::fgRejectNoITS = kTRUE;
Double_t      AliAODSimpleEvent::fgPtLimit4PID = 1000000.0;
Double_t      AliAODSimpleEvent::fgProbThreshold = 0.1;
Bool_t        AliAODSimpleEvent::fgUsePerfectPID = kTRUE;

//--------------------------------------------------------------------------------------------------------
AliAODSimpleEvent::AliAODSimpleEvent(Int_t evnum, Int_t ntracks) 
//
// Default (and unique) constructor.
// Initializes pointers to NULL ans makes ready for a default selection.
//
{
	if (!fgTracks) {
		fgTracks = new TClonesArray("AliAODSimpleTrack", ntracks);
		if (fgTracks->CanBypassStreamer()) fgTracks->BypassStreamer();
	}
	fTracks = fgTracks;
	
	fEventNumber = evnum;
	SetPrimaryVertex(0.0, 0.0, 0.0);
		
	Int_t i;
	for (i = 0; i < kArrayIndexes; i++) {
		fRef[i] = new TRefArray;
		if (i < kParticleTypes) fgPrior[i] = 1.0;
	}
	
	fgDefMass[kElectronPlus] = fgDefMass[kElectronMinus] = 0.000511;
	fgDefMass[kMuonPlus] = fgDefMass[kMuonMinus] = 0.105658357;
	fgDefMass[kPionPlus] = fgDefMass[kPionMinus] = 0.13957018;
	fgDefMass[kKaonPlus] = fgDefMass[kKaonMinus] = 0.493677;
	fgDefMass[kProtonPlus] = fgDefMass[kProtonMinus] = 0.938272;
	
	fgDefPDG[kElectron] = 11;
	fgDefPDG[kMuon] = 13;
	fgDefPDG[kPion] = 211;
	fgDefPDG[kKaon] = 321;
	fgDefPDG[kProton] = 2212;
}
//--------------------------------------------------------------------------------------------------------
AliAODSimpleEvent::~AliAODSimpleEvent()
//
// Destructor.
// Clears the TObjArray
//
{
	Clear();
	
	Int_t i;
	for (i = 0; i < kArrayIndexes; i++) {
		delete fRef[i];
		fRef[i] = 0;
	}
}
//--------------------------------------------------------------------------------------------------------
void AliAODSimpleEvent::Clear(Option_t * /*option*/)
//
// Clears list of tracks and references
//
{
	fTracks->Clear(); 
	
	Int_t i;
	for (i = 0; i < kArrayIndexes; i++) fRef[i]->Delete();
}
//--------------------------------------------------------------------------------------------------------
void AliAODSimpleEvent::Reset(Option_t * /*option*/)
//
// Deletes all static pointer objects.
//
{
	delete fgTracks; 
	fgTracks = 0;
}
//--------------------------------------------------------------------------------------------------------
TArrayI AliAODSimpleEvent::DoSelection(AliESD *esd, TObject *partList)
// 
// Performs particle selection.
// Second argument must contain a kind of collection object where
// simulated particle information is stored.
// It can be a TClonesArray (in case of miniesd) of AliStack, while
// in any other case (or if it is NULL), simulation infos are not stored.
// Of course, particles are selected only if their corresponding collection
// has been initialized with AliAODSimpleEvent::InitCollection().
//
{
	// output
	TArrayI counts(kArrayIndexes);
	
	// clear
	Clear();
	
	// if 'list' argument is NULL, perfect PID cannot be done
	if (!partList && fgUsePerfectPID) {
		Error("DoSelection", "Cannot perform 'perfect' PID if list of particles is not passed!");
		return counts;
	}
	
	// check class name of 'list' argument
	Bool_t isAliStack = kFALSE, isTClonesArray = kFALSE;
	TString className(partList->ClassName());
	isAliStack = ( (!className.CompareTo("AliStack", TString::kExact)) );
	isTClonesArray = ( (!className.CompareTo("TClonesArray", TString::kExact)) );
	if (!isAliStack && !isTClonesArray) {
		Error("Select", "Passed list type '%s' is not managed by this class", className.Data());
		return counts;
	}
	
	// define variables for all useful track data
	Int_t i, ntracks, sign, index, label, motherLabel, motherPDG, assignedPDG;
	Bool_t fillArray[kArrayIndexes];
	Double_t p[3], v[3];
	TArrayD pid(kParticleTypes);
	for (i = 0; i < kParticleTypes; i++) pid[i] = 0.0;
	
	// get number of tracks
	ntracks = esd->GetNumberOfTracks();
	
	// check array size
	//if (ntracks != fTracks->GetSize()) {
	//	fTracks->Expand(ntracks);
	//}
	
	// loop on tracks
	for (index = 0; index < ntracks; index++) {
		
		// get ESD track
		AliESDtrack *esdTrack = esd->GetTrack(index);
		
		// if not all 3 passes of tracking have been performed, then the track is rejected
		if (fgRejectNoITS) {
			if (!(esdTrack->GetStatus() & AliESDtrack::kITSrefit)) continue;
		}
		
		// get vertex and momentum and check on their values
		// an EXATCLY zero value means that the track has something strange
		esdTrack->GetPxPyPz(p);
		esdTrack->GetXYZ(v);
		sign = esdTrack->GetSign();
		if (v[0] == 0.0 || v[1] == 0.0 || v[2] == 0.0) continue;
		if (p[0] == 0.0 || p[1] == 0.0 || p[2] == 0.0) continue;
		
		// get kinematics particle related to the track
		TParticle *part = 0;
		label = TMath::Abs(esdTrack->GetLabel());
		motherLabel = -1;
		motherPDG = 0;
		if (partList) {
			if (isTClonesArray) part = (TParticle*)((TClonesArray*)partList)->At(label);
			if (isAliStack) part = (TParticle*)((AliStack*)partList)->Particle(label);
			motherLabel = part->GetFirstMother();
			motherPDG = 0;
			if (motherLabel >= 0) {
				TParticle *mum = 0;
				if (isTClonesArray) mum = (TParticle*)((TClonesArray*)partList)->At(motherLabel);
				if (isAliStack) mum = (TParticle*)((AliStack*)partList)->Particle(motherLabel);
				if (mum) motherPDG = mum->GetPdgCode();
			}
		}
		
		// create new track object and initialize datamembers (except mass)
		TClonesArray &tracks = *fTracks;
		AliAODSimpleTrack *track = new(tracks[index]) AliAODSimpleTrack(kParticleTypes);
		track->Index() = index;
		track->Label() = label;
		track->MotherLabel() = motherLabel;
		track->MotherPDG() = motherPDG;
		track->SetPID(pid);
		track->SetPxPyPz(p[0], p[1], p[2]);
		track->SetVxVyVz(v[0], v[1], v[2]);
		
		// identify particle and store it according to the selected choice
		for (i = 0; i < kArrayIndexes; i++) fillArray[i] = kFALSE;
		if (fgUsePerfectPID) {
			// 'Perfect' PID: take the true PDG code
			assignedPDG = part->GetPdgCode();
			i = GetArrayIndex(assignedPDG);
			fillArray[i] = kTRUE;
		}
		else {
			// 'Real' PID: invoke Bayesian PID
			Double_t esdpid[kParticleTypes];
			esdTrack->GetESDpid(esdpid);
			pid.Adopt(kParticleTypes, esdpid);
			assignedPDG = BayesianPID(pid) * sign;
			if (track->Get3Momentum().Perp() <= fgPtLimit4PID) {
				i = GetArrayIndex(assignedPDG);
				fillArray[i] = kTRUE;
			}
			else {
				if (sign > 0) {
					for (i = 0; i < kArrayIndexes; i += 2) fillArray[i] = kTRUE;
				}
				else {
					for (i = 1; i < kArrayIndexes; i += 2) fillArray[i] = kTRUE;
				}
			}
		}
		
		// add the track to all collections to which it has to be put
		// they can be many if the PID info is not taken into account
		for (i = 0; i < kArrayIndexes; i++) {
			if (!fillArray[i]) continue;
			track->SetMass(fgDefMass[i]);
			fRef[i]->Add(track);
			counts[i]++;
		}
	} // end of loop over tracks
	
	return counts;
}
//--------------------------------------------------------------------------------------------------------
TRefArray* AliAODSimpleEvent::GetTracks(EArrayIndex type)
//
// Returns the particle collection specified in argument
//
{
	if (type < 0 || type >= kArrayIndexes) return NULL;
	return fRef[(Int_t)type];
}
//--------------------------------------------------------------------------------------------------------
inline AliAODSimpleEvent::EArrayIndex AliAODSimpleEvent::GetArrayIndex(Int_t pdgCode)
//
// [PRIVATE]
// Returns the array index of the pdg code specified
//
{
	switch (pdgCode) {
		case    -11: return kElectronPlus;
		case     11: return kElectronMinus;
		case    -13: return kMuonPlus;
		case     13: return kMuonMinus;
		case    211: return kPionPlus;
		case   -211: return kPionMinus;
		case    321: return kKaonPlus;
		case   -321: return kKaonMinus;
		case   2212: return kProtonPlus;
		case  -2212: return kProtonMinus;
		default:     return kNoIndex;
	}
}
//--------------------------------------------------------------------------------------------------------
Int_t AliAODSimpleEvent::BayesianPID(TArrayD &pid)
//
// [PRIVATE]
// Combines the PID weights recovered from ESD object and prior probabilities,
// and creates an array of 5 elements containing the final probabilities to
// identify the particle as one of the 5 particle types taken into account
// (electron, muon, pion, kaon, proton).
// If the largest array element corresponds to a pion or kaon, and it is larger than the threshold,
// the function returns their PDG code, otherwise it returns 0.
//
{
	// step 1 - compute the normalization factor
	Int_t i;
	Double_t sum = 0.0;
	for (Int_t i = 0; i < kParticleTypes; i++) {
		sum += fgPrior[i] * pid[i];
	}
	if (sum <= 0.0) return 0;
	
	// step 2 - normalize PID weights by the relative prior probability
	for (Int_t i = 0; i < kParticleTypes; i++) {
		pid[i] *= fgPrior[i] / sum;
	}
	
	// step 3 - find the largest array element and assign to related particle
	Int_t imax = 0;
	for (i = 1; i < kParticleTypes; i++) {
		if (pid[i] > pid[imax]) imax = i;
	}
	
	// step 4 - check the value of the largest element against the threshold
	if (pid[imax] >= fgProbThreshold) {
		return fgDefPDG[imax];
	}
	else {
		return 0;
	}
}
//--------------------------------------------------------------------------------------------------------
